# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import Optional

from openworld.sdk.core.model.error import Error
from openworld.sdk.core.model.exception.openworld import OpenWorldException


class OpenWorldServiceException(OpenWorldException):
    def __init__(self, message: str, cause: Optional[BaseException] = None):
        super().__init__(message, cause)

    @staticmethod
    def of(error: Error, error_code: HTTPStatus):
        return OpenWorldServiceException(message=f"[{error_code.value}] {error}")


class OpenWorldAuthException(OpenWorldServiceException):
    def __init__(self, error_code: HTTPStatus, message: str):
        super().__init__(message=f"[{error_code.value}] {message}")



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests

from openworld.sdk.core.constant import constant, log


def response_log(response: requests.Response):
    headers: dict = response.headers
    headers_log: str = constant.EMPTY_STRING

    for key, value in headers.items():
        headers_log += f"\t\t{key}: {value}\n"

    if not headers_log.endswith("\n"):
        headers_log += "\n"

    result: str = (
        "\nResponse:\n" + log.HTTP_HEADERS_LOG_MESSAGE_TEMPLATE.format(headers_log) + log.HTTP_BODY_LOG_MESSAGE_TEMPLATE.format("\t\t" + response.text + "\n")
    )

    return result


def request_log(headers: dict, body: str, endpoint: str, method: str, response: requests.Response):
    headers_log: str = constant.EMPTY_STRING

    for key, value in headers.items():
        headers_log += f"\t\t{key}: {value}\n"

    if not headers_log.endswith("\n"):
        headers_log += "\n"

    result = f"\nRequest: {endpoint}\n" f"Method: {method.upper()}\n" + log.HTTP_HEADERS_LOG_MESSAGE_TEMPLATE.format(
        headers_log
    ) + log.HTTP_BODY_LOG_MESSAGE_TEMPLATE.format("\t\t" + body + "\n") + response_log(response)

    return result


def filter_credentials(data: dict):
    new_data = dict()
    filter_keys = ["key", "secret", "username", "password"]
    for key, value in data.items():
        for word in filter_keys:
            if word.lower() in key.lower():
                new_data[key] = "<-- omitted -->"
                break
            new_data[key] = value
    return new_data



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class Location(Enum):
    HEADER = "HEADER"
    PATH = "PATH"
    QUERY = "QUERY"
    BODY = "BODY"


@dataclass_json
@dataclass
class ErrorCause:
    type: str
    detail: str
    location: Location
    name: str
    value: str


@dataclass_json
@dataclass
class Error:
    type: str
    detail: str
    causes: Optional[list[ErrorCause]] = None



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from openworld.sdk.core.constant import message
from openworld.sdk.core.constant.constant import EMPTY_STRING
from openworld.sdk.core.constant.url import AUTH_ENDPOINT
from openworld.sdk.core.model.authentication import Credentials
from openworld.sdk.core.model.exception import client as client_exception

DEFAULT_CREDENTIALS = Credentials(key=EMPTY_STRING, secret=EMPTY_STRING)


@dataclass
class AuthConfig:
    def __init__(
        self,
        credentials: Credentials = DEFAULT_CREDENTIALS,
        auth_endpoint: str = AUTH_ENDPOINT,
    ):
        r"""Holds authentication config data.

        :param credentials: Client's credentials.
        :param auth_endpoint: URL to use as a base for oauth token requests, has a default value of [ACCESS_TOKEN] if
                               not provided.
        """
        self.__credentials: Credentials = credentials
        self.__auth_endpoint: str = auth_endpoint

        self.__post_init__()

    def __post_init__(self):
        missing = list()

        for attribute in [
            ("Key", self.__credentials.key),
            ("Secret", self.__credentials.secret),
            ("Auth_Endpoint", self.__auth_endpoint),
        ]:

            if not attribute[1]:
                missing.append(attribute[0])

        if len(missing):
            raise client_exception.OpenWorldConfigurationException(message=message.NONE_VALUE_NOT_ALLOWED_FOR_MESSAGE_TEMPLATE.format(f"{missing}"))

    @property
    def credentials(self) -> Credentials:
        return self.__credentials

    @property
    def auth_endpoint(self) -> str:
        return self.__auth_endpoint



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

BEARER: str = "Bearer "

AUTHORIZATION: str = "Authorization"

GRANT_TYPE: str = "grant_type"

CLIENT_CREDENTIALS: str = "client_credentials"

CONTENT_TYPE: str = "Content-type"

JSON_CONTENT_TYPE: str = "application/json"

ACCEPT: str = "Accept"

TRANSACTION_ID: str = "transaction-id"

USER_AGENT: str = "User-agent"

OPENWORLD_SDK_PYTHON: str = "open-world-sdk-python/"

API_REQUEST: dict = {CONTENT_TYPE: JSON_CONTENT_TYPE, ACCEPT: JSON_CONTENT_TYPE}



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from http import HTTPStatus

from requests import Response, post
from requests.auth import HTTPBasicAuth

from openworld.sdk.core.constant import body as body_constant
from openworld.sdk.core.constant import log as log_constant
from openworld.sdk.core.constant import url as url_constant
from openworld.sdk.core.constant.constant import OK_STATUS_CODES_RANGE
from openworld.sdk.core.constant.message import UNABLE_TO_AUTHENTICATE
from openworld.sdk.core.model.authentication import Credentials, Token
from openworld.sdk.core.model.exception import service as service_exception
from openworld.sdk.core.util import log as log_util

LOG = logging.getLogger(__name__)


class _AuthClient:
    def __init__(self, credentials: Credentials, auth_endpoint: str = url_constant.AUTH_ENDPOINT):
        r"""Manages user authentication process.

        :param credentials: Client key and secret pair
        """
        self.__credentials: Credentials = credentials
        self.__auth_endpoint: str = auth_endpoint

        token_response = self.__retrieve_token(auth_endpoint=self.__auth_endpoint).json()
        self.__token: Token = Token(token_response)

    def __retrieve_token(self, auth_endpoint: str = url_constant.AUTH_ENDPOINT) -> Response:
        LOG.info(log_constant.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(log_constant.TOKEN_RENEWAL_IN_PROCESS))

        auth_method = HTTPBasicAuth(username=self.__credentials.key, password=self.__credentials.secret)

        response = post(url=auth_endpoint, auth=auth_method, data=body_constant.TOKEN_REQUEST)

        if response.status_code not in OK_STATUS_CODES_RANGE:
            raise service_exception.OpenWorldAuthException(
                message=UNABLE_TO_AUTHENTICATE,
                error_code=HTTPStatus(response.status_code),
            )

        request_log_message = log_util.request_log(
            headers=log_util.filter_credentials(auth_method.__dict__),
            body=str(body_constant.TOKEN_REQUEST),
            endpoint=auth_endpoint,
            method="post",
            response=response,
        )

        LOG.info(log_constant.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(request_log_message))
        return response

    def refresh_token(self) -> None:
        r"""Refreshes access token.

        :return: the new token.
        :rtype: requests.Response
        """
        if not self.__token.is_about_expired():
            return

        if self.__token.lock.acquire(block=True):
            if self.__token.is_about_expired():
                LOG.info(log_constant.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(log_constant.TOKEN_EXPIRED))
                response = self.__retrieve_token(auth_endpoint=self.__auth_endpoint)
                self.__token.update(data=response.json())
                LOG.info(log_constant.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(log_constant.TOKEN_RENEWAL_SUCCESSFUL))
            self.__token.lock.release()

    @property
    def access_token(self) -> str:
        r"""Gets the access token value.

        :return: the access token value.
        :rtype: str
        """
        return self.__token.access_token

    def is_token_expired(self):
        return self.__token.is_expired()

    def is_token_about_expired(self):
        return self.__token.is_about_expired()



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime
from enum import Enum
from http import HTTPStatus
from typing import Any, Optional

import requests

from openworld.sdk.core.client.auth_client import _AuthClient
from openworld.sdk.core.configuration.client_config import ClientConfig
from openworld.sdk.core.constant import header as header_constant
from openworld.sdk.core.constant import log as log_constant
from openworld.sdk.core.constant.constant import OK_STATUS_CODES_RANGE
from openworld.sdk.core.model.authentication import HttpBearerAuth
from openworld.sdk.core.model.error import Error
from openworld.sdk.core.model.exception import service as service_exception
from openworld.sdk.core.util import log as log_util

LOG = logging.getLogger(__name__)


class ApiClient:
    def __init__(self, config: ClientConfig):
        r"""Sends requests to API.

        :param config: Client Configuration Wrapper
        """
        self.__auth_client: _AuthClient = _AuthClient(
            credentials=config.auth_config.credentials,
            auth_endpoint=config.auth_config.auth_endpoint,
        )

        self.endpoint = config.endpoint

    @staticmethod
    def __build_response(response: requests.Response, response_model):
        if response.status_code not in OK_STATUS_CODES_RANGE:
            raise service_exception.OpenWorldServiceException.of(
                error=Error.from_json(response.json()),
                error_code=HTTPStatus(response.status_code),
            )

        if response_model is None:
            return None

        return response_model.from_dict(response.json())

    def call(
        self,
        method: str,
        url: str,
        obj: Any = None,
        request_headers: Optional[dict] = None,
        response_model: Optional[Any] = None,
    ) -> Any:
        r"""Sends HTTP request to API.

        :param method: Http request method
        :param obj: Object that holds request data
        :param response_model: Model to fetch the response data into
        :param url: URL used to send the request
        :param request_headers: Headers of the request

        :return: response as object
        :rtype: Any
        """
        self.__auth_client.refresh_token()

        request_headers = ApiClient.__fill_request_headers(request_headers)

        auth_bearer = HttpBearerAuth(access_token=self.__auth_client.access_token)
        request_body = dict()

        if not obj:
            response = requests.request(
                method=method.upper(),
                url=url,
                headers=request_headers,
                auth=auth_bearer,
            )
        else:
            request_body = obj.to_json(default=ApiClient.__serialization_helper)
            response = requests.request(
                method=method.upper(),
                url=url,
                headers=request_headers,
                data=request_body,
                auth=auth_bearer,
            )

        request_log_message = log_util.request_log(
            headers=request_headers,
            body=str(request_body),
            endpoint=url,
            method=method,
            response=response,
        )

        LOG.info(log_constant.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(request_log_message))

        result = ApiClient.__build_response(response=response, response_model=response_model)
        return result

    @staticmethod
    def __serialization_helper(obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        if isinstance(obj, Enum):
            return obj.value

    @staticmethod
    def __fill_request_headers(request_headers: dict):
        if not request_headers:
            request_headers = dict()

        request_header_keys = request_headers.keys()
        for key in header_constant.API_REQUEST.keys():
            if key in request_header_keys:
                continue

            request_headers[key] = header_constant.API_REQUEST[key]

        return request_headers



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import Optional

from openworld.sdk.core.configuration.auth_config import AuthConfig
from openworld.sdk.core.constant import message, url
from openworld.sdk.core.model.authentication import Credentials
from openworld.sdk.core.model.exception import client as client_exception


@dataclass
class ClientConfig:
    __auth_config: AuthConfig = None

    def __init__(
        self,
        key: str,
        secret: str,
        endpoint: Optional[str] = url.ENDPOINT,
        auth_endpoint: Optional[str] = url.AUTH_ENDPOINT,
    ):
        r"""SDK Client Configurations Holder.

        :param key: The API key to use for authentication.
        :param secret: The API secret to use for authentication.
        :param endpoint: An optional API endpoint to use for requests.
        :param auth_endpoint: An optional API endpoint to use for authentication.
        """
        self.__auth_config = AuthConfig(Credentials(key, secret), auth_endpoint)
        self.__endpoint = endpoint

        self.__post_init__()

    def __post_init__(self):
        if not self.__endpoint:
            raise client_exception.OpenWorldConfigurationException(message.NONE_VALUE_NOT_ALLOWED_FOR_MESSAGE_TEMPLATE.format(self.__endpoint))

    @property
    def auth_config(self) -> AuthConfig:
        return self.__auth_config

    @property
    def endpoint(self) -> str:
        return self.__endpoint



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser

from setuptools import setup


def parse_version():
    config_parser = configparser.ConfigParser()
    config_parser.read("./core.config")
    return config_parser["sdk-properties"]["version"]


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="openworld-sdk-python-core",
    version=f"{parse_version()}",
    package_data={"config_files": ["core.config"]},
    packages=[
        "openworld.sdk.core",
        "openworld.sdk.core.configuration",
        "openworld.sdk.core.client",
        "openworld.sdk.core.constant",
        "openworld.sdk.core.model",
        "openworld.sdk.core.model.exception",
        "openworld.sdk.core.util",
    ],
    package_dir={"openworld.sdk.core": "."},
    license="Apache License, Version 2.0",
    author="Expedia Group",
    author_email="oss@expediagroup.com",
    url="https://github.com/ExpediaGroup/openworld-sdk-python",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8",
    install_requires=["dataclasses_json", "uri", "requests", "python-dateutil"],
    description="Open World SDK Core Library for Python",
    long_description=readme(),
    long_description_content_type="text/markdown",
)



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
from dataclasses import dataclass
from multiprocessing import Lock
from typing import Optional

import requests
from dataclasses_json import dataclass_json
from requests.auth import AuthBase

from openworld.sdk.core.constant import header, log
from openworld.sdk.core.constant.constant import REFRESH_TOKEN_TIME_GAP_IN_SECONDS

LOG = logging.getLogger(__name__)


@dataclass
class Credentials:
    """A pair of client key and secret."""

    key: str
    secret: str


@dataclass_json
@dataclass
class _TokenResponse:
    """A model of an API response."""

    access_token: str
    expires_in: int
    scope: str
    token_type: str
    id_token: Optional[str] = None
    refresh_token: Optional[str] = None


@dataclass
class Token:
    def __init__(self, data: dict):
        r"""Represents a token model.

        :param data: token data
        """
        self.__token = _TokenResponse.from_dict(data)
        self.lock = Lock()
        self.__expiration_time = datetime.datetime.now() + datetime.timedelta(seconds=self.__token.expires_in)

        LOG.info(log.OPENWORLD_LOG_MESSAGE_TEMPLATE.format(log.NEW_TOKEN_EXPIRATION_TEMPLATE.format(str(self.__token.expires_in))))

    @property
    def refresh_token(self) -> str:
        return self.__token.refresh_token

    @property
    def access_token(self) -> str:
        return self.__token.access_token

    @property
    def id_token(self) -> str:
        return self.__token.id_token

    def is_expired(self):
        return datetime.datetime.now() >= self.__expiration_time

    def is_about_expired(self):
        return datetime.datetime.now() + datetime.timedelta(seconds=REFRESH_TOKEN_TIME_GAP_IN_SECONDS) >= self.__expiration_time

    def update(self, data: dict):
        self.__token = _TokenResponse.from_dict(data)
        self.__expiration_time = datetime.datetime.now() + datetime.timedelta(seconds=self.__token.expires_in)


@dataclass
class HttpBearerAuth(AuthBase):
    r"""Holds Bearer access token."""

    __access_token: str

    def __init__(self, access_token):
        self.__access_token = access_token

    def __call__(self, request: requests.Request = None) -> requests.Request:
        request.headers[header.AUTHORIZATION] = header.BEARER + self.__access_token
        return request

    def __str__(self) -> str:
        return header.BEARER + self.__access_token



# Copyright 2022 Expedia, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from textwrap import dedent

TOKEN_RENEWAL_IN_PROCESS: str = "Renewing token"

TOKEN_RENEWAL_SUCCESSFUL: str = "Token renewal successful"

TOKEN_EXPIRED: str = "Token expired or is about to expire, request will wait until token is renewed"

OPENWORLD_LOG_MESSAGE_TEMPLATE: str = "ExpediaSDK: {0}"

UNSUCCESSFUL_RESPONSE_MESSAGE_TEMPLATE: str = "Unsuccessful response [{0}]"

NEW_TOKEN_EXPIRATION_TEMPLATE: str = "New token expires in {0} seconds"

HTTP_HEADERS_LOG_MESSAGE_TEMPLATE: str = dedent(
    """\tHeaders:
    \t--- BEGIN ---
    {0}
    \t--- END ---
    """
)

HTTP_BODY_LOG_MESSAGE_TEMPLATE: str = dedent(
    """
    \tBody:
    \t--- BEGIN ---
    {0}
    \t--- END ---
    """
)
